package org.rykov.gismeteo.idea;

import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.util.*;
import com.intellij.openapi.options.Configurable;
import com.intellij.openapi.options.ConfigurationException;
import com.intellij.util.net.HttpConfigurable;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.Nls;
import org.jetbrains.annotations.Nullable;
import org.rykov.gismeteo.common.*;
import org.rykov.gismeteo.ui.Utils;
import org.rykov.gismeteo.ui.Resources;
import org.rykov.gismeteo.ui.Config;
import org.rykov.gismeteo.sax.MMWeatherHandler;
import org.rykov.gismeteo.sax.DocumentHandler;
import org.jdom.Element;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import javax.swing.*;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.util.*;
import java.util.Timer;
import java.net.URL;
import java.net.URLConnection;
import java.text.MessageFormat;
import java.beans.XMLEncoder;
import java.beans.XMLDecoder;
import java.awt.*;

/**
 * This class defines application component.
 * @author rykov
 * @created 21.10.2008 15:32:23
 */
public class GismeteoApplicationComponent extends Observable implements ApplicationComponent,
        NamedJDOMExternalizable, Configurable {

    /** The town index. */
    private String index = Config.get("gismeteo.index");

    /** The weather forecast. */
    private MMWeather weather;

    /** The forecast request last modifed header. */
    private long lastModified;

    /** The logger. */
    private Logger LOGGER = Logger.getLogger(getClass());

    /** The index field name. */
    private static final String INDEX = "index";

    /** The weather field name. */
    private static final String WEATHER = "weather";

    /** The last modified field name. */
    private static final String LAST_MODIFIED = "lastModified";

    /** The saved configuration filename. */
    private static final String CONFIG_NAME = "gismeteo.plugin";

    /** The configuration panel. */
    private GismeteoConfiguration conf;

    /** The path to town data. */
    private static final String PATH = "/org/rykov/gismeteo/resources/towns/";

    /** The countries. */
    private Document countries;

    /** The towns. */
    private Document towns;

    /**
     * Inits component.
     */
    public void initComponent() {
        loadTowns();
        loadWeather();

        Timer timer = new Timer();
        timer.scheduleAtFixedRate(new TimerTask() {
            public void run() {
                EventQueue.invokeLater(new Runnable() {
                    public void run() {
                        loadWeather();
                    }
                });
            }
        }, Utils.HOUR - (System.currentTimeMillis() % Utils.HOUR), Utils.HOUR);
    }

    /**
     * Disposes component.
     */
    public void disposeComponent() {}

    @NotNull
    public String getComponentName() {
        return "GismeteoApplicationComponent";
    }

    /**
     * Loads towns.
     */
    private void loadTowns() {
        try {
            LOGGER.info("Load towns");
            SAXParserFactory factory = SAXParserFactory.newInstance();
            SAXParser parser = factory.newSAXParser();
            DocumentHandler countriesHandler = new DocumentHandler();
            parser.parse(getClass().getResourceAsStream(PATH + "clist.xml"), countriesHandler);
            countries = countriesHandler.getDocument();
            DocumentHandler townsHandler = new DocumentHandler();
            parser.parse(getClass().getResourceAsStream(PATH + "tlist.xml"), townsHandler);
            towns = townsHandler.getDocument();
        } catch (SAXException e) {
            LOGGER.info("SAX exception", e);
        } catch (ParserConfigurationException e) {
            LOGGER.info("SAX exception", e);
        } catch (IOException e) {
            LOGGER.info("SAX exception", e);
        }
    }

    /**
     * Loads the weather forecast.
     */
    synchronized private void loadWeather() {
        InputStream is = null;
        try {
            LOGGER.info("Try load weather for " + index);
            SAXParserFactory factory = SAXParserFactory.newInstance();
            SAXParser parser = factory.newSAXParser();
            HttpConfigurable.getInstance().setAuthenticator();
            String url = MessageFormat.format(Config.get("gismeteo.xml"), index);
            URLConnection connection = new URL(url).openConnection();
            long lm = connection.getLastModified();
            LOGGER.info("Get last modified " + new Date(lm));
            if (weather == null || !weather.getReport().getTown().getIndex().equals(index) || lm > lastModified) {
                LOGGER.info("Load weather");
                MMWeatherHandler handler = new MMWeatherHandler();
                is = connection.getInputStream();
                parser.parse(is, handler);
                weather = handler.getWeather();
                lastModified = lm;
                setChanged();
                notifyObservers();
            } else {
                LOGGER.info("Keep old weather");
            }
        } catch (SAXException e) {
            LOGGER.info("SAX exception", e);
        } catch (ParserConfigurationException e) {
            LOGGER.info("SAX exception", e);
        } catch (IOException e) {
            LOGGER.info("SAX exception", e);
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    LOGGER.error("Can't close request input stream");
                }
            }
        }
    }

    synchronized public MMWeather getWeather() {
        return weather;
    }

    /**
     * Gets the town name by town index.
     * @param index the town index
     * @return the town name, null if not found
     */
    public String getTownName(String index) {
        for (TownInfo ti : towns.getTowns()) {
            if (ti.getIndex().equals(index)) {
                return ti.getName_en();
            }
        }
        return null;
    }

    public void readExternal(Element element) throws InvalidDataException {
        index = JDOMExternalizer.readString(element, INDEX);
        ByteArrayInputStream stream = new ByteArrayInputStream(JDOMExternalizer.readString(element, WEATHER).getBytes());
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(MMWeather.class.getClassLoader());
        XMLDecoder decoder = new XMLDecoder(stream);
        weather = (MMWeather) decoder.readObject();
        decoder.close();
        Thread.currentThread().setContextClassLoader(cl);
        try {
            lastModified = Long.parseLong(JDOMExternalizer.readString(element, LAST_MODIFIED));
        } catch (NumberFormatException e) {
            LOGGER.warn("Can't parse last modified field");
        }
    }

    public void writeExternal(Element element) throws WriteExternalException {
        JDOMExternalizer.write(element, INDEX, index);
        ByteArrayOutputStream stream = new ByteArrayOutputStream();
        XMLEncoder encoder = new XMLEncoder(stream);
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(MMWeather.class.getClassLoader());
        encoder.writeObject(weather);
        encoder.close();
        Thread.currentThread().setContextClassLoader(cl);
        JDOMExternalizer.write(element, WEATHER, stream.toString());
        JDOMExternalizer.write(element, LAST_MODIFIED, "" + lastModified);
    }

    @NonNls
    public String getExternalFileName() {
        return CONFIG_NAME;
    }

    @Nls
    public String getDisplayName() {
        return Resources.get("gismeteo.title");
    }

    @Nullable
    public Icon getIcon() {
        return null;
    }

    @Nullable
    @NonNls
    public String getHelpTopic() {
        return null;
    }

    public JComponent createComponent() {
        if (conf == null) {
            conf = new GismeteoConfiguration(countries, towns);
        }
        return conf;
    }

    public boolean isModified() {
        return !index.equals(conf.getIndex());
    }

    public void apply() throws ConfigurationException {
        if (conf.getIndex() == null || conf.getIndex().length() != Utils.INDEX_LENGHT) {
            throw new ConfigurationException(Resources.get("conf.bad.index"));
        }
        index = conf.getIndex();
        loadWeather();
    }

    public void reset() {
        conf.setIndex(index);
    }

    public void disposeUIResources() {
        conf = null;
    }
}
